#include <stdint.h>
#include <memory.h>
#include <debug.h>
#include <multiboot.h>
#include <musl-glue.h>

uint64_t kernel_P4;

void memory_init()
{
    kernel_P4 = (uint64_t)&BootP4;

    uint64_t start, end;
    uint32_t type, i = 0;
    while(!multiboot_get_memory_area(i++, &start, &end, &type))
    {
        for(uint64_t p = start; p < end; p += PAGE_SIZE)
        {
            uint64_t vaddr = (uint64_t)P2V(p);
            uint64_t page = vmm_get_page(kernel_P4, vaddr);
            if(page == (uint64_t)-1 || !(page & PAGE_PRESENT))
            {
                uint16_t flags = PAGE_WRITE | PAGE_PRESENT;
                vmm_set_page(kernel_P4, vaddr, p, flags);
            }

            if(type != MMAP_FREE)
                continue;
            if(p >= V2P(&kernel_start) && p < V2P(&kernel_end))
                continue;
            if(multiboot_page_used((uintptr_t)P2V(p)))
                continue;

            pmm_free(p);
        }
    }
}